/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>

#define PORT 65432
#define BUFSIZE 128
#define MAX_SEQ_ID 10
#define ACK "ACK"
#define TIMEOUT_SEC 3
#define N 4

int do_write(int sock, int id)
{
  char buf[BUFSIZE];

  memset(buf, 0, sizeof(buf));
  sprintf(buf, "%d", id);
  printf("Message to server: %s\n", buf);
  return write(sock, buf, sizeof(buf));
}

int main(void)
{
  int sock;
  fd_set set;
  struct sockaddr_in client;
  char buf[BUFSIZE];
  struct timeval timeout;
  int ret;
  int i, j, n;
  int total = 0;
  int current_index;

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  memset(&client, 0, sizeof(client));
  client.sin_family = AF_INET;
  client.sin_port = htons(PORT);
  client.sin_addr.s_addr = inet_addr("127.0.0.1");

  if (connect(sock, (struct sockaddr *) &client, sizeof(client)) == -1) {
    perror("connect");
    goto end;
  }

  while (total <= MAX_SEQ_ID) {
    current_index = total;

    for (i = current_index; i < current_index + N && i <= MAX_SEQ_ID; i++) {
      if (do_write(sock, i) == -1) {
        perror("write");
        goto end;
      }

      usleep(1000);
    }

    for (i = current_index; i < current_index + N; i++) {
resend:
      FD_ZERO(&set);
      FD_SET(sock, &set);
      timeout.tv_sec = TIMEOUT_SEC;
      timeout.tv_usec = 0;

      if ((ret = select(sock + 1, &set, NULL, NULL, &timeout)) == -1) {
        perror("select");
        goto end;
      } else if (ret == 0) {
        printf("Timeout for message: %d\n", i);

        if (do_write(sock, i) == -1) {
          perror("write");
          goto end;
        }

        usleep(1000);
        goto resend;
      } else {
        if ((n = read(sock, buf, sizeof(buf))) == -1) {
          perror("read");
          goto end;
        } else if (n == 0) {
          continue;
        }

        buf[n] = '\0';

        printf("Message from server: %s\n", buf);

        if (strncmp(buf, ACK, strlen(ACK)) != 0) {
          j = atoi(buf + strlen(ACK));
          printf("Corrupt message ACK%d\n", j);

          if (do_write(sock, j) == -1) {
            perror("write");
            goto end;
          }

          usleep(1000);
          goto resend;
        } else {
          total++;
        }
      }
    }
  }

end:

  if (close(sock) == -1) {
    perror("close");
    exit(2);
  }

  return 0;
}
