/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <fcntl.h>

#define PORT 65432
#define BUFSIZE 128
#define BACKLOG 1024
#define ACK "ACK"
#define NAK "NAK"
#define MAX_SEQ_ID 10
#define DISCARD_ID 5

bool is_fault(void)
{
  int d = rand() % 2;
  return d > 0;
}

int main(void)
{
  int s_sock, c_sock;
  struct sockaddr_in server, other;
  socklen_t other_size;
  char buf[BUFSIZE];
  bool flag = true;
  int i, n;
  int isfault;
  char buf2[BUFSIZE];

  // srand(time(NULL));
  srand(0);

  if ((s_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  memset(&server, 0, sizeof(server));
  memset(&other, 0, sizeof(other));
  server.sin_family = AF_INET;
  server.sin_port = htons(PORT);
  server.sin_addr.s_addr = INADDR_ANY;

  if (bind(s_sock, (struct sockaddr *) &server, sizeof(server)) == -1) {
    perror("bind");
    goto end;
  }

  if (listen(s_sock, BACKLOG) == -1) {
    perror("listen");
    goto end;
  }

  other_size = sizeof(other);

  if ((c_sock = accept(s_sock, (struct sockaddr *)&other, &other_size)) == -1) {
    perror("accept");
    goto end;
  }

  i = 0;

  while (i <= MAX_SEQ_ID) {
    memset(buf, 0, sizeof(buf));
    memset(buf2, 0, sizeof(buf2));

    if (i == DISCARD_ID && flag == true) {
      printf("i = %d: [Simulation loss]\n", i);
      flag = false;

      if (read(c_sock, buf, sizeof(buf)) == -1) {
        perror("read");
        goto end2;
      }

      continue;
    }

    if ((n = read(c_sock, buf, sizeof(buf))) == -1) {
      perror("read");
      goto end2;
    }

    isfault = is_fault();
    printf("Message from client: %s\n", buf);
    printf("Corruption status: %d\n", isfault);

    if (isfault) {
      strcpy(buf2, NAK);
    } else {
      strcpy(buf2, ACK);
      i++;
    }

    strcat(buf2, buf);
    printf("Message to client: %s\n", buf2);

    if (write(c_sock, buf2, sizeof(buf2)) == -1) {
      perror("write");
      goto end2;
    }
  }

end2:

  if (close(c_sock) == -1) {
    perror("close");
    exit(2);
  }

end:

  if (close(s_sock) == -1) {
    perror("close");
    exit(3);
  }

  return 0;
}
